/* Philips PCF50606 Power Management Unit (PMU) driver
 * (C) 2006 by Harald Welte <laforge@gnumonks.org>
 *
 * This driver is a monster ;) It provides the following features
 * - voltage control for a dozen different voltage domains
 * - charging control for main and backup battery
 * - rtc / alarm
 * - watchdog
 * - adc driver (hw_sensors like)
 */


/* public in-kernel pcf50606 api */
enum pcf50606_regulator_id {
	PCF50606_REGULATOR_DCD,
	PCF50606_REGULATOR_DCDE,
	PCF50606_REGULATOR_DCUD,
	PCF50606_REGULATOR_D1REG,
	PCF50606_REGULATOR_D2REG,
	PCF50606_REGULATOR_D3REG,
	PCF50606_REGULATOR_LPREG,
	__NUM_PCF50606_REGULATORS
}
struct pcf50606_data;
extern void pcf50606_go_standby(struct pcf50606_data *pcf);
extern int pcf50606_voltage_set(struct pcf50606_data *pcf,
				enum pcf50606_regulator_id reg, 
				unsigned int millivolts);

struct pcf50606_platform_data {
	unsigned int initial_voltage[__NUM_PCF50606_REGULATORS];
	
	struct {
		u_int8_t mbcc3; /* charger voltage / current */
	} charger;
};

#include <linux/module.h>
#include <linux/init.h>
#include <linux/i2c.h>

static unsigned short normal_i2c[] = { 0x08, I2C_CLIENT_END };

I2C_CLIENT_INSMOD_1(pcf50606);

struct pcf50606_data {
	struct i2c_client client;
	spinlock_t lock;
};

static inline int __reg_write(struct pcf50606_data *pcf, u_int8_t reg, u_int8_t val)
{
	return i2c_smbus_write_byte_data(&pcf->client, reg, val); 	
}

static int reg_write(struct pcf50606_data *pcf, u_int8_t reg, u_int8_t val)
{
	int ret;

	spin_lock_irq(&pcf->lock);
	ret = __reg_write(pcf, reg, val);
	spin_unlock_irq(&pcf->lock);

	return ret;
}

static inline int32_t __reg_read(struct pcf50606_data *pcf, u_int8_t reg)
{
	return i2c_smbus_read_byte_data(&pcf->client, reg);
}

static int32_t reg_read(struct pcf50606_data *pcf, u_int8_t reg, u_int8_t *data)
{
	int ret;

	spin_lock_irq(&pcf->lock);
	ret = __reg_read(pcf, reg, data);
	spin_unlock_irq(&pcf->lock);

	return ret;
}

static int reg_set_bit_mask(struct pcf50606_data *pcf, 
			    u_int8_t reg, u_int8_t val, u_int8_t mask)
{
	int ret;
	u_int8_t tmp;
	spin_lock_irq(&pcf->lock);

	tmp = __reg_read(pcf, reg);
	tmp &= ~mask;
	tmp |= val;
	ret = __reg_write(pcf, reg, tmp);

	spin_unlock_irq(&pcf->lock);

	return ret;
}


static int reg_clear_bits(struct pcf50606_data *pcf, u_int8_t reg, u_int8_t val)
{
	int ret;
	u_int8_t tmp;

	spin_lock_irq(&pcf->lock);

	tmp = __reg_read(pcf, reg);
	tmp &= ~val;
	ret = __reg_write(pcf, reg, tmp);

	spin_unlock_irq(&pcf->lock);

	return ret;
}

static u_int8_t dcudc_voltage(unsigned int millivolts)
{
	if (millivolts < 900)
		return 0;
	if (millivolts > 5500)
		return 0x1f;
	if (millivolts <= 3300) {
		millivolts -= 900;
		return millivolts/300;
	}
	if (millivolts < 4000)
		return 0x0f;
	else {
		millivolts -= 4000;
		return millivolts/100;
	}
}

static u_int8_t dcdec_voltage(unsigned int millivolts)
{
	if (millivolts < 900)
		return 0;
	else if (millivolts > 3300)
		return 0x0f;
	
	millivolts -= 900;
	return millivolts/300;
}

static u_int8_t dcdc_voltage(unsigned int millivolts)
{
	if (millivolts < 900)
		return 0;
	else if (millivolts > 3600)
		return 0x1f;
	
	millivolts -= 900;
	return millivolts/25;
}

static u_int8_t dx_voltage(unsigned int millivolts)
{

}

static u_int8_t lp_voltage(unsigned int millivolts)
{

}

int pcf50606_voltage_set(struct pcf50606_data *pcf,
			 enum pcf50606_regulator_id reg, unsigned int millivolts)
{
	u_int8_t volt_bits;
	u_int8_t regnr;
	int rc;

	if (reg >= __NUM_PCF50606_REGULATORS)
		return -EINVAL;

	switch (reg) {
	case PCF50606_REGULATOR_DCD:
		volt_bits = dcdc_voltage(millivolts);
		rc = reg_set_bit_mask(pcf, PCF50606_REG_DCDC1, 0x1f,
				      volt_bits);
		break;
	case PCF50606_REGULATOR_DCDE:
		volt_bits = dcdec_voltage(millivolts);
		rc = reg_set_bit_mask(pcf, PCF50606_REG_DCDEC1, 0x0f,
				      volt_bits);
		break;
	case PCF50606_REGULATOR_DCUD:
		volt_bits = dcudc_voltage(millivolts);
		rc = reg_set_bit_mask(pcf, PCF50606_REG_DCUDC1, 0x1f,
				      volt_bits);
		break;
	case PCF50606_REGULATOR_D1REG:
	case PCF50606_REGULATOR_D2REG:
	case PCF50606_REGULATOR_D3REG:
		regnr = PCF50606_REG_D1REGC1 + (reg - PCF50606_REGULATOR_D1REG);
		volt_bits = dx_voltage(millivolts);
		rc = reg_set_bit_mask(pcf, regnr, 0x1f, volt_bits);
		break;
	case PCF50606_REGULATOR_LPREG:
		volt_bits = lp_voltage(millivolts);
		rc = reg_set_bit_mask(pcf, PCF50606_REG_LPREGC1, 0x1f,
				      volt_bits);
		break;
	}

	return rc;
}
EXPORT_SYMBOL(pcf50606_voltage_set);


/* go into 'STANDBY' mode, i.e. power off the main CPU and peripherals */
void pcf50606_go_standby(struct pcf50606_data *pcf)
{
	reg_write(pcf, PCF50606_REG_OOCC1, PCF50606_OOCC1_GOSTDBY);
}
EXPORT_SYMBOL(pcf50606_go_standby);



static irqreturn_t pcf50606_irq(int irq, void *_pcf, struct pt_regs *regs)
{
	struct pcf50606_data *pcf = _pcf;
	u_int8_t val;

	val = __reg_read(pcf, PCF50606_REG_INT1);
	DEBUGP("INT1=0x%02x ", val);
	if (val & PCF50606_INT1_ONKEYR) {
	}
	if (val & PCF50606_INT1_ONKEYF) {
	}
	if (val & PCF50606_INT1_ONKEY1S) {
	}
	/* FIXME: EXTONR, EXTONF, SECOND */
	if (val & PCF50606_INT1_ALARM) {
	}

	val = __reg_read(pcf, PCF50606_REG_INT2);
	DEBUGP("INT2=0x%02x ", val);

	if (val & PCF50606_INT2_CHGINS) {
	}
	if (val & PCF50606_INT2_CHGRM) {
	}
	if (val & PCF50606_INT2_CHGFOK) {
	}
	if (val & PCF50606_INT2_CHGERR) {
	}
	if (val & PCF50606_INT2_CHGFRDY) {
	}
	if (val & PCF50606_INT2_CHGPROT) {
	}
	if (val & PCF50606_INT2_CHGWD10S) {
	}
	if (val & PCF50606_INT2_CHGWDEXP) {
	}

	val = __reg_read(pcf, PCF50606_REG_INT3);
	DEBUGP("INT3=0x%02x\n", val);

	if (val & PCF50606_INT3_ADCRDY) {
	}
	/* FIXME: ACDINS, ACDREM, TSCPRES */
	if (val & PCF50606_LOWBBAT) {
	}
	if (val & PCF50606_HIGHTMP) {
	}

	/* FIXME: clear interrupt sources, return IRQ_HANDLED conditionally */

	return IRQ_HANDLED;
}



static struct i2c_driver pcf50606_driver;

static int pcf50606_detect(struct i2c_adapter *adapter, int address, int kind)
{
	struct pcf50606_data *data;

	if (!(data = kzalloc(sizeof(*data), GFP_KERNEL)))
		return -ENOMEM;
	
	i2c_set_clientdata(&data->client, data);
	data->client.addr = address;
	data->client.adapter = adapter;
	data->client.driver = &pcf50606_driver;
	data->client.flags = 0;

	/* now we try to detect the chip */

}

static int pcf50606_attach_adapter(struct i2c_adapter *adapter)
{
	return i2c_probe(adapter, &addr_data, &pcf50606_detect);
}


static struct i2c_driver pcf50606_driver = {
	.driver = {
		.name	= "pcf50606",
	},
	.id		= I2C_DRIVERID_PCF50606,
	.attach_adapter	= &pcf50606_attach_adapter,
	.detach_client	= &pcf50606_detach_client,
};

static int __init pcf50606_init(void)
{
	return i2c_add_driver(&pcf50606_driver);
}

static void __exit pcf50606_exit(void)
{
	i2c_del_driver(&pcf50606_driver);
}


MODULE_AUTHOR("Harald Welte <laforge@gnumonks.org>");
MODULE_LICENSE("GPL");

module_init(pcf50606_init);
module_exit(pcf50606_exit);

