/*
 *  Backlight Driver for FIC GTA01
 *
 *  based on corg_cl.c, Copyright (c) 2004-2006 Richard Purdie
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/mutex.h>
#include <linux/fb.h>
#include <linux/backlight.h>
#include <asm/arch/gta01.h>

static int gta01bl_intensity;
static DEFINE_MUTEX(bl_mutex);
static struct backlight_properties gta01bl_data;
static struct backlight_device *gta01_backlight_device;
static struct gta01bl_machinfo *bl_machinfo;

static unsigned long gta01bl_flags;
#define CORGIBL_SUSPENDED     0x01
#define CORGIBL_BATTLOW       0x02

static int gta01bl_send_intensity(struct backlight_device *bd)
{
	int intensity = bd->props->brightness;

	if (bd->props->power != FB_BLANK_UNBLANK)
		intensity = 0;
	if (bd->props->fb_blank != FB_BLANK_UNBLANK)
		intensity = 0;
	if (gta01bl_flags & CORGIBL_SUSPENDED)
		intensity = 0;
	if (gta01bl_flags & CORGIBL_BATTLOW)
		intensity &= bl_machinfo->limit_mask;

 	mutex_lock(&bl_mutex);
	bl_machinfo->set_bl_intensity(intensity);
	mutex_unlock(&bl_mutex);

	gta01bl_intensity = intensity;

	return 0;
}

#ifdef CONFIG_PM
static int gta01bl_suspend(struct platform_device *dev, pm_message_t state)
{
	gta01bl_flags |= CORGIBL_SUSPENDED;
	gta01bl_send_intensity(gta01_backlight_device);
	return 0;
}

static int gta01bl_resume(struct platform_device *dev)
{
	gta01bl_flags &= ~CORGIBL_SUSPENDED;
	gta01bl_send_intensity(gta01_backlight_device);
	return 0;
}
#else
#define gta01bl_suspend	NULL
#define gta01bl_resume	NULL
#endif

static int gta01bl_get_intensity(struct backlight_device *bd)
{
	return gta01bl_intensity;
}

static int gta01bl_set_intensity(struct backlight_device *bd)
{
	gta01bl_send_intensity(gta01_backlight_device);
	return 0;
}

/*
 * Called when the battery is low to limit the backlight intensity.
 * If limit==0 clear any limit, otherwise limit the intensity
 */
void gta01bl_limit_intensity(int limit)
{
	if (limit)
		gta01bl_flags |= CORGIBL_BATTLOW;
	else
		gta01bl_flags &= ~CORGIBL_BATTLOW;
	gta01bl_send_intensity(gta01_backlight_device);
}
EXPORT_SYMBOL(gta01bl_limit_intensity);


static struct backlight_properties gta01bl_data = {
	.owner          = THIS_MODULE,
	.get_brightness = gta01bl_get_intensity,
	.update_status  = gta01bl_set_intensity,
};

static int __init gta01bl_probe(struct platform_device *pdev)
{
	struct gta01bl_machinfo *machinfo = pdev->dev.platform_data;

	bl_machinfo = machinfo;
	gta01bl_data.max_brightness = machinfo->max_intensity;
	if (!machinfo->limit_mask)
		machinfo->limit_mask = -1;

	gta01_backlight_device = backlight_device_register ("gta01-bl",
		NULL, &gta01bl_data);
	if (IS_ERR (gta01_backlight_device))
		return PTR_ERR (gta01_backlight_device);

	gta0ibl_data.power = FB_BLANK_UNBLANK;
	gta01bl_data.brightness = machinfo->default_intensity;
	gta01bl_send_intensity(gta01_backlight_device);

	/* somehow use s3c_device_timer0 for PWM */

	printk("GTA01 Backlight Driver Initialized.\n");
	return 0;
}

static int gta01bl_remove(struct platform_device *dev)
{
	backlight_device_unregister(gta01_backlight_device);

	printk("GTA01 Backlight Driver Unloaded\n");
	return 0;
}

static struct platform_driver gta01bl_driver = {
	.probe		= gta01bl_probe,
	.remove		= gta01bl_remove,
	.suspend	= gta01bl_suspend,
	.resume		= gta01bl_resume,
	.driver		= {
		.name	= "gta01-bl",
	},
};

static int __init gta01bl_init(void)
{
	return platform_driver_register(&gta01bl_driver);
}

static void __exit gta01bl_exit(void)
{
	platform_driver_unregister(&gta01bl_driver);
}

module_init(gta01bl_init);
module_exit(gta01bl_exit);

MODULE_AUTHOR("Harald Welte <laforge@gnumonks.org>");
MODULE_DESCRIPTION("GTA01 Backlight Driver");
MODULE_LICENSE("GPLv2");
